// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

import 'dart:io';
import 'dart:math';

/// Generate functions like this one:
///
///     R Function<T1> Function<T2> c12(R fn(T1, T2)) {
///       return (T1 a1) => (T2 a2) => fn(a1, a2);
///     }
///
void main(List<String> arguments) {
  final n = int.parse(arguments.single);
  final file = new File('lib/src/generated.dart');
  final sink = file.openWrite();
  sink.write('part of partial;\n\n');
  sink.write('// THIS FILE IS AUTO-GENERATED, DO NOT EDIT!\n\n');

  // Write each function on a single line.
  for (var l = 2; l <= n; l++) {
    final source = new List<int>.generate(l, (i) => i + 1);

    // Number of combinations is l!
    for (var i = 0; i < fac(l); i++) {
      // Get l position numbers in the range (1, 2, ... l).
      final indices = new List<int>(l);
      var prevStep = 0;
      for (var j = 0; j < l; j++) {
        final range = j + 1;
        final step = max(1, j * prevStep);
        prevStep = step;
        indices[j] = (i / step).floor() % range;
      }

      // Generate pattern from source.
      final copy = source.toList();
      final v = new List<int>();
      for (final j in indices.reversed) {
        v.add(copy.removeAt(j));
      }

      // Write function definition.
      final s = source;
      sink.write('R ${v.reversed.map((i) => 'Function(T$i)').join(' ')} c'
          '${v.map((i) => '$i').join()}<R,${s.map((i) => 'T$i').join(',')}>'
          '(R fn(${s.map((i) => 'T$i a$i').join(',')})) '
          '{ return ${v.map((i) => '(T$i a$i)').join(' => ')} => fn('
          '${s.map((i) => 'a$i').join(',')}); }\n');
    }
  }

  sink.close();
}

int fac(int n) {
  var product = 1;
  for (var i = 1; i <= n; i++) {
    product *= i;
  }
  return product;
}
